from __future__ import annotations

from dataclasses import dataclass


class TanukiError(Exception):
    pass


class TanukiTimeoutError(TanukiError):
    pass


class TanukiConnectionError(TanukiError):
    pass


class TanukiParcelError(TanukiError):
    pass


@dataclass(frozen=True, slots=True)
class RemoteErrorPayload:
    type: str
    message: str
    traceback: str | None = None


class TanukiRemoteError(TanukiError):
    def __init__(self, payload: RemoteErrorPayload):
        self.payload = payload
        msg = payload.message
        if payload.traceback:
            msg = f"{msg}\n\nRemote traceback:\n{payload.traceback}"
        super().__init__(msg)



from __future__ import annotations

import hashlib
import mimetypes
import os
import shutil
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from .exceptions import TanukiParcelError


@dataclass(frozen=True, slots=True)
class S3Config:
    bucket: str | None
    endpoint_url: str | None = None
    region_name: str | None = None
    prefix: str = "tanu"
    addressing_style: str | None = None  # "path" or "virtual"
    access_key_id: str | None = None
    secret_access_key: str | None = None
    session_token: str | None = None

    @classmethod
    def from_env(cls) -> "S3Config":
        region = os.getenv("TANU_S3_REGION") or os.getenv("AWS_DEFAULT_REGION") or os.getenv("AWS_REGION")
        prefix = os.getenv("TANU_S3_PREFIX", "tanu")
        addressing_style = os.getenv("TANU_S3_ADDRESSING_STYLE") or None
        return cls(
            bucket=os.getenv("TANU_S3_BUCKET") or None,
            endpoint_url=os.getenv("TANU_S3_ENDPOINT_URL") or None,
            region_name=region or None,
            prefix=prefix,
            addressing_style=addressing_style,
            access_key_id=os.getenv("TANU_S3_ACCESS_KEY_ID") or None,
            secret_access_key=os.getenv("TANU_S3_SECRET_ACCESS_KEY") or None,
            session_token=os.getenv("TANU_S3_SESSION_TOKEN") or None,
        )


def _sha256_file(path: Path) -> str:
    h = hashlib.sha256()
    with path.open("rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            h.update(chunk)
    return h.hexdigest()


def _safe_guess_content_type(filename: str) -> str:
    guess, _ = mimetypes.guess_type(filename)
    return guess or "application/octet-stream"


def _normalize_prefix(prefix: str) -> str:
    return prefix.strip().strip("/")


def _gen_key(prefix: str, filename: str) -> str:
    ext = Path(filename).suffix
    uid = uuid.uuid4().hex
    p = _normalize_prefix(prefix)
    if p:
        return f"{p}/{uid}{ext}"
    return f"{uid}{ext}"


class S3ParcelStore:
    def __init__(self, *, config: S3Config) -> None:
        self._config = config
        self._client = None

    @property
    def config(self) -> S3Config:
        return self._config

    def _get_client(self):  # type: ignore[no-untyped-def]
        if self._client is not None:
            return self._client

        try:
            import boto3
            from botocore.config import Config as BotocoreConfig
        except ModuleNotFoundError as e:
            raise TanukiParcelError("boto3 is required for S3 Parcel support (install: pip install tanu[s3])") from e

        botocore_cfg: BotocoreConfig | None = None
        if self._config.addressing_style:
            botocore_cfg = BotocoreConfig(s3={"addressing_style": self._config.addressing_style})
        else:
            botocore_cfg = BotocoreConfig()

        kwargs: dict[str, Any] = {}
        if self._config.endpoint_url:
            kwargs["endpoint_url"] = self._config.endpoint_url
        if self._config.region_name:
            kwargs["region_name"] = self._config.region_name
        if self._config.access_key_id:
            kwargs["aws_access_key_id"] = self._config.access_key_id
        if self._config.secret_access_key:
            kwargs["aws_secret_access_key"] = self._config.secret_access_key
        if self._config.session_token:
            kwargs["aws_session_token"] = self._config.session_token
        kwargs["config"] = botocore_cfg

        self._client = boto3.client("s3", **kwargs)
        return self._client

    def put_file(
        self,
        path: Path,
        *,
        bucket: str | None = None,
        key: str | None = None,
        filename: str | None = None,
        content_type: str | None = None,
    ) -> dict[str, Any]:
        target_bucket = bucket or self._config.bucket
        if not target_bucket:
            raise TanukiParcelError("S3 bucket is not configured (set TANU_S3_BUCKET)")

        filename = filename or path.name
        content_type = content_type or _safe_guess_content_type(filename)
        target_key = key or _gen_key(self._config.prefix, filename)

        size = int(path.stat().st_size)
        sha256 = _sha256_file(path)

        extra_args: dict[str, Any] = {"ContentType": content_type}
        try:
            self._get_client().upload_file(str(path), target_bucket, target_key, ExtraArgs=extra_args)
        except Exception as e:
            raise TanukiParcelError(f"failed to upload to S3: {e}") from e

        uri = f"s3://{target_bucket}/{target_key}"
        return {
            "uri": uri,
            "bucket": target_bucket,
            "key": target_key,
            "filename": filename,
            "content_type": content_type,
            "size": size,
            "sha256": sha256,
        }

    def download_file(self, *, bucket: str, key: str, dst: Path) -> None:
        dst.parent.mkdir(parents=True, exist_ok=True)
        try:
            self._get_client().download_file(bucket, key, str(dst))
        except Exception as e:
            raise TanukiParcelError(f"failed to download from S3: {e}") from e


_DEFAULT_S3_CONFIG: S3Config | None = None
_DEFAULT_S3_STORE: S3ParcelStore | None = None


def get_default_s3_store() -> S3ParcelStore:
    global _DEFAULT_S3_CONFIG, _DEFAULT_S3_STORE
    cfg = S3Config.from_env()
    if _DEFAULT_S3_STORE is not None and _DEFAULT_S3_CONFIG == cfg:
        return _DEFAULT_S3_STORE
    _DEFAULT_S3_CONFIG = cfg
    _DEFAULT_S3_STORE = S3ParcelStore(config=cfg)
    return _DEFAULT_S3_STORE


class Parcel:
    """
    File reference transferred via S3-compatible object storage.

    - Send: Parcel("/path/to/file")  # uploads during RPC serialization
    - Receive: Parcel(...)           # created from metadata in response/request
    - Download: parcel.save("path")  # downloads from object storage
    """

    def __init__(
        self,
        path: str | os.PathLike[str] | None = None,
        *,
        uri: str | None = None,
        bucket: str | None = None,
        key: str | None = None,
        filename: str | None = None,
        content_type: str | None = None,
        size: int | None = None,
        sha256: str | None = None,
    ) -> None:
        self._local_path: Path | None = Path(path) if path is not None else None
        self.uri = uri
        self.bucket = bucket
        self.key = key
        self.filename = filename or (self._local_path.name if self._local_path else None)
        self.content_type = content_type
        self.size = size
        self.sha256 = sha256

    def _ensure_remote(self) -> dict[str, Any]:
        if self.bucket and self.key:
            uri = self.uri or f"s3://{self.bucket}/{self.key}"
            return {
                "uri": uri,
                "bucket": self.bucket,
                "key": self.key,
                "filename": self.filename,
                "content_type": self.content_type,
                "size": self.size,
                "sha256": self.sha256,
            }

        if self._local_path is None:
            raise TanukiParcelError("Parcel has no local path and no remote reference")
        if not self._local_path.exists():
            raise TanukiParcelError(f"Parcel file not found: {self._local_path}")

        meta = get_default_s3_store().put_file(
            self._local_path,
            filename=self.filename or self._local_path.name,
            content_type=self.content_type or _safe_guess_content_type(self._local_path.name),
        )
        self.uri = str(meta.get("uri") or self.uri)
        self.bucket = str(meta.get("bucket") or "")
        self.key = str(meta.get("key") or "")
        self.filename = str(meta.get("filename") or self.filename or self._local_path.name)
        self.content_type = str(meta.get("content_type") or self.content_type or _safe_guess_content_type(self.filename))
        self.size = int(meta.get("size") or self.size or self._local_path.stat().st_size)
        self.sha256 = str(meta.get("sha256") or self.sha256 or "")
        return meta

    def save(self, filename: str | os.PathLike[str] | None = None) -> str:
        if not self.bucket or not self.key:
            if self._local_path is None:
                raise TanukiParcelError("Parcel has no remote reference (bucket/key) and no local file")
            dst = Path(filename) if filename is not None else Path(self.filename or self._local_path.name)
            if dst.is_dir():
                dst = dst / (self.filename or self._local_path.name)
            dst.parent.mkdir(parents=True, exist_ok=True)
            if dst.resolve() != self._local_path.resolve():
                shutil.copy2(self._local_path, dst)
            return str(dst)

        dst = Path(filename) if filename is not None else Path(self.filename or Path(self.key).name)
        if dst.is_dir():
            dst = dst / (self.filename or Path(self.key).name)
        get_default_s3_store().download_file(bucket=self.bucket, key=self.key, dst=dst)
        return str(dst)



"""
Backward-compatible public module.

Prefer importing from `tanu` directly:
    from tanu import Tanuki, TanukiWorker
"""

from .client import Tanuki
from .parcel import Parcel
from .worker import TanukiWorker

__all__ = ["Parcel", "Tanuki", "TanukiWorker"]



from __future__ import annotations

from dataclasses import dataclass, field
import logging
import threading
import time
from collections.abc import Callable
from inspect import Parameter, Signature, signature
from typing import Any

import pika

from .config import RabbitMQConfig
from .protocol import build_error_response, build_ok_response, build_progress
from .rabbitmq import connect_blocking, declare_rpc_topology, declare_worker_queue
from .utils.object_encoder_decoder import decode_json, encode_json

logger = logging.getLogger("tanu.worker")

Handler = Callable[..., Any]
ProgressCallback = Callable[..., Any]


@dataclass(frozen=True, slots=True)
class ProgressContext:
    worker_name: str
    method: str
    args: list[Any]
    kwargs: dict[str, Any]
    started_at: float
    elapsed: float


ProgressProvider = Callable[[ProgressContext], Any]


@dataclass(slots=True)
class _InFlightRequest:
    method: str
    args: list[Any]
    kwargs: dict[str, Any]
    props: pika.BasicProperties
    delivery_tag: int | None
    started_at: float
    last_progress_at: float
    progress_interval: float
    progress_provider: ProgressProvider | None
    done: threading.Event = field(default_factory=threading.Event)
    response: dict[str, Any] | None = None
    client_gone: bool = False


def _normalize_progress_callback(fn: ProgressCallback) -> ProgressProvider:
    try:
        sig = signature(fn)
    except Exception:
        return lambda _ctx: fn()  # type: ignore[misc]

    if not sig.parameters:
        return lambda _ctx: fn()  # type: ignore[misc]
    return lambda ctx: fn(ctx)  # type: ignore[misc]


class TanukiWorker:
    def __init__(
        self,
        worker_name: str,
        config: RabbitMQConfig | None = None,
        *,
        worker_id: str | None = None,
    ) -> None:
        self._worker_name = worker_name
        self._config = config or RabbitMQConfig()
        self._worker_id = worker_id or f"tanuki-worker:{worker_name}"

        _, self._realm = self._config.split_worker_name(worker_name)
        self._request_exchange = self._config.request_exchange_name(worker_name)
        self._reply_exchange = self._config.reply_exchange_name(worker_name)
        self._routing_key = self._config.request_routing_key(worker_name)

        self._handlers: dict[str, Handler] = {}
        self._progress_handlers: dict[str, ProgressProvider] = {}
        self._progress_intervals: dict[str, float] = {}
        self.register("__tanuki_help__", self._tanuki_help)

        self._connection: pika.BlockingConnection | None = None
        self._consume_ch: pika.adapters.blocking_connection.BlockingChannel | None = None
        self._publish_ch: pika.adapters.blocking_connection.BlockingChannel | None = None
        self._queue_name: str | None = None
        self._stop_requested = False

    @property
    def config(self) -> RabbitMQConfig:
        return self._config

    @property
    def worker_name(self) -> str:
        return self._worker_name

    def handler(
        self,
        method: str,
        *,
        progress: ProgressCallback | None = None,
        progress_interval: float | None = None,
    ) -> Callable[[Handler], Handler]:
        def _decorator(fn: Handler) -> Handler:
            self._handlers[method] = fn
            if progress is not None:
                self.set_progress(method, progress, interval=progress_interval)
            return fn

        return _decorator

    def command(
        self,
        name: str,
        *,
        progress: ProgressCallback | None = None,
        progress_interval: float | None = None,
    ) -> Callable[[Handler], Handler]:
        return self.handler(name, progress=progress, progress_interval=progress_interval)

    def register(
        self,
        method: str,
        fn: Handler,
        *,
        progress: ProgressCallback | None = None,
        progress_interval: float | None = None,
    ) -> None:
        self._handlers[method] = fn
        if progress is not None:
            self.set_progress(method, progress, interval=progress_interval)

    def progress(self, method: str, *, interval: float | None = None) -> Callable[[ProgressCallback], ProgressCallback]:
        def _decorator(fn: ProgressCallback) -> ProgressCallback:
            self.set_progress(method, fn, interval=interval)
            return fn

        return _decorator

    def set_progress(self, method: str, fn: ProgressCallback | None, *, interval: float | None = None) -> None:
        if fn is None:
            self._progress_handlers.pop(method, None)
            self._progress_intervals.pop(method, None)
            return

        self._progress_handlers[method] = _normalize_progress_callback(fn)
        if interval is not None:
            self._progress_intervals[method] = float(interval)

    def _tanuki_help(self) -> dict[str, Any]:
        import inspect

        def _ann_to_str(ann: Any) -> str | None:
            if ann is inspect._empty:
                return None
            if isinstance(ann, type):
                return ann.__name__
            return str(ann)

        def _param_to_dict(p: Parameter) -> dict[str, Any]:
            return {
                "name": p.name,
                "kind": p.kind.name,
                "annotation": _ann_to_str(p.annotation),
                "has_default": p.default is not inspect._empty,
                "default": None if p.default is inspect._empty else repr(p.default),
            }

        commands: list[dict[str, Any]] = []
        for name, fn in sorted(self._handlers.items(), key=lambda kv: kv[0]):
            if name.startswith("__tanuki_"):
                continue
            try:
                sig: Signature = inspect.signature(fn)
                sig_str = f"{name}{sig}"
                params = [_param_to_dict(p) for p in sig.parameters.values()]
                return_ann = _ann_to_str(sig.return_annotation)
            except Exception:
                sig_str = name
                params = []
                return_ann = None

            doc = getattr(fn, "__doc__", None) or None
            if isinstance(doc, str):
                doc = doc.strip().splitlines()[0] if doc.strip() else None

            commands.append(
                {
                    "name": name,
                    "signature": sig_str,
                    "params": params,
                    "return": return_ann,
                    "doc": doc,
                }
            )

        return {
            "worker": self._worker_name,
            "realm": self._realm,
            "queue": self._config.request_queue_name(self._worker_name),
            "request_exchange": self._request_exchange,
            "reply_exchange": self._reply_exchange,
            "commands": commands,
        }

    def close(self) -> None:
        if self._consume_ch and self._consume_ch.is_open:
            try:
                self._consume_ch.close()
            except Exception:
                logger.debug("ignore close(consume_ch) error", exc_info=True)
        if self._publish_ch and self._publish_ch.is_open:
            try:
                self._publish_ch.close()
            except Exception:
                logger.debug("ignore close(publish_ch) error", exc_info=True)
        if self._connection and self._connection.is_open:
            try:
                self._connection.close()
            except Exception:
                logger.debug("ignore close(connection) error", exc_info=True)
        self._connection = None
        self._consume_ch = None
        self._publish_ch = None
        self._queue_name = None

    def stop(self) -> None:
        self._stop_requested = True
        self.close()

    def purge(self) -> int:
        self._ensure_connected()
        assert self._consume_ch and self._queue_name
        result = self._consume_ch.queue_purge(queue=self._queue_name)
        return int(result.method.message_count)

    def _ensure_connected(self) -> None:
        if self._connection and self._connection.is_open and self._consume_ch and self._publish_ch and self._queue_name:
            return

        self.close()

        self._connection = connect_blocking(self._config, connection_name=self._worker_id, max_retries=0)
        self._consume_ch = self._connection.channel()
        self._publish_ch = self._connection.channel()

        declare_rpc_topology(self._consume_ch, self._config, worker_name=self._worker_name)
        self._queue_name = declare_worker_queue(self._consume_ch, self._config, worker_name=self._worker_name)
        declare_rpc_topology(self._publish_ch, self._config, worker_name=self._worker_name)

        self._publish_ch.confirm_delivery()
        self._consume_ch.basic_qos(prefetch_count=self._config.prefetch_count)

    def _publish_reply(self, props: pika.BasicProperties, body: bytes, *, msg_type: str, warn_unroutable: bool) -> bool:
        if not props.reply_to:
            return False
        assert self._publish_ch is not None
        try:
            self._publish_ch.basic_publish(
                exchange=self._reply_exchange,
                routing_key=props.reply_to,
                body=body,
                properties=pika.BasicProperties(
                    content_type="application/json",
                    correlation_id=props.correlation_id,
                    app_id=self._worker_id,
                    type=msg_type,
                ),
                mandatory=True,
            )
            return True
        except (pika.exceptions.UnroutableError, pika.exceptions.NackError):
            if warn_unroutable:
                logger.warning("%s unroutable (client likely gone): reply_to=%s", msg_type, props.reply_to)
            else:
                logger.debug("%s unroutable (client likely gone): reply_to=%s", msg_type, props.reply_to)
            return False
        except pika.exceptions.AMQPError:
            logger.exception("failed to publish %s", msg_type)
            return False

    def _ack_if_needed(self, delivery_tag: int | None) -> None:
        if self._config.worker_auto_ack or delivery_tag is None:
            return
        assert self._consume_ch is not None
        try:
            self._consume_ch.basic_ack(delivery_tag=delivery_tag)
        except Exception:
            logger.debug("ignore ack error", exc_info=True)

    def _start_inflight(self, delivery_tag: int | None, props: pika.BasicProperties, body: bytes) -> _InFlightRequest | None:
        try:
            req = decode_json(body)
            method_name = str(req.get("method"))
            args = req.get("args") or []
            kwargs = req.get("kwargs") or {}
            if not isinstance(args, list) or not isinstance(kwargs, dict):
                raise ValueError("invalid request payload")

            fn = self._handlers.get(method_name)
            if fn is None:
                raise KeyError(f"unknown method: {method_name}")
        except KeyError as e:
            logger.warning(str(e))
            resp = build_error_response(e, include_traceback=self._config.include_traceback_in_error)
            self._publish_reply(props, encode_json(resp), msg_type="tanuki.response", warn_unroutable=True)
            self._ack_if_needed(delivery_tag)
            return None
        except Exception as e:
            logger.exception("failed to decode request")
            resp = build_error_response(e, include_traceback=self._config.include_traceback_in_error)
            self._publish_reply(props, encode_json(resp), msg_type="tanuki.response", warn_unroutable=True)
            self._ack_if_needed(delivery_tag)
            return None

        progress_provider = self._progress_handlers.get(method_name)
        progress_interval = float(self._progress_intervals.get(method_name, self._config.progress_interval))

        inflight = _InFlightRequest(
            method=method_name,
            args=args,
            kwargs=kwargs,
            props=props,
            delivery_tag=delivery_tag,
            started_at=time.monotonic(),
            last_progress_at=0.0,
            progress_interval=progress_interval,
            progress_provider=progress_provider,
        )

        def _run_handler() -> None:
            try:
                result = fn(*args, **kwargs)
                resp = build_ok_response(result)
            except KeyError as e:
                logger.warning(str(e))
                resp = build_error_response(e, include_traceback=self._config.include_traceback_in_error)
            except Exception as e:
                logger.exception("handler error")
                resp = build_error_response(e, include_traceback=self._config.include_traceback_in_error)
            inflight.response = resp
            inflight.done.set()

        threading.Thread(target=_run_handler, name=f"tanuki:{method_name}", daemon=True).start()
        return inflight

    def _maybe_send_progress(self, inflight: _InFlightRequest) -> None:
        if inflight.client_gone or inflight.progress_interval <= 0:
            return
        now = time.monotonic()
        if now - inflight.last_progress_at < inflight.progress_interval:
            return

        elapsed = now - inflight.started_at
        ctx = ProgressContext(
            worker_name=self._worker_name,
            method=inflight.method,
            args=inflight.args,
            kwargs=inflight.kwargs,
            started_at=inflight.started_at,
            elapsed=elapsed,
        )

        progress_payload: Any = {"status": "running"}
        if inflight.progress_provider is not None:
            try:
                progress_payload = inflight.progress_provider(ctx)
            except Exception:
                logger.exception("progress callback error; fallback to auto progress")
                inflight.progress_provider = None

        progress_msg = build_progress(progress_payload, elapsed=elapsed)
        try:
            progress_body = encode_json(progress_msg)
        except Exception:
            logger.exception("failed to serialize progress; fallback to auto progress")
            try:
                progress_body = encode_json(build_progress({"status": "running"}, elapsed=elapsed))
            except Exception:
                return

        if not inflight.props.reply_to:
            inflight.last_progress_at = now
            return

        ok = self._publish_reply(inflight.props, progress_body, msg_type="tanuki.progress", warn_unroutable=False)
        if not ok:
            inflight.client_gone = True
        inflight.last_progress_at = now

    def _finish_inflight(self, inflight: _InFlightRequest) -> None:
        resp = inflight.response or build_error_response(
            RuntimeError("missing response"),
            include_traceback=self._config.include_traceback_in_error,
        )
        try:
            resp_body = encode_json(resp)
        except Exception as e:
            logger.exception("failed to serialize response")
            resp_body = encode_json(build_error_response(e, include_traceback=self._config.include_traceback_in_error))

        if inflight.props.reply_to:
            self._publish_reply(inflight.props, resp_body, msg_type="tanuki.response", warn_unroutable=True)
        else:
            logger.debug("no reply_to set; dropping response")
        self._ack_if_needed(inflight.delivery_tag)

    def run(self, *, configure_logging: bool = True, log_level: int | str = "INFO") -> None:
        if configure_logging:
            _ensure_default_logging(log_level)

        startup_logged = False
        inflight: _InFlightRequest | None = None
        self._stop_requested = False

        while not self._stop_requested:
            try:
                self._ensure_connected()
                assert self._connection and self._consume_ch and self._queue_name

                if not startup_logged:
                    commands = sorted(k for k in self._handlers.keys() if not k.startswith("__tanuki_"))
                    cmd_str = ",".join(commands) if commands else "-"
                    realm = self._realm or "local"
                    if self._realm is None:
                        logger.info("TanukiWorker ready: %s (local) queue=%s cmds=%s", self._worker_name, self._queue_name, cmd_str)
                    else:
                        logger.info(
                            "TanukiWorker ready: %s (realm=%s) queue=%s req=%s rep=%s cmds=%s",
                            self._worker_name,
                            realm,
                            self._queue_name,
                            self._request_exchange,
                            self._reply_exchange,
                            cmd_str,
                        )
                    logger.debug(
                        "TanukiWorker connection: rabbitmq=%s:%s vhost=%s",
                        self._config.host,
                        self._config.port,
                        self._config.virtual_host,
                    )
                    startup_logged = True
                else:
                    logger.info("TanukiWorker reconnected: %s (realm=%s) queue=%s", self._worker_name, self._realm or "local", self._queue_name)

                while not self._stop_requested:
                    if inflight is None:
                        method_frame, props, body = self._consume_ch.basic_get(queue=self._queue_name, auto_ack=self._config.worker_auto_ack)
                        if method_frame is None:
                            self._connection.process_data_events(time_limit=1.0)
                            continue
                        delivery_tag = None if self._config.worker_auto_ack else int(method_frame.delivery_tag)
                        inflight = self._start_inflight(delivery_tag, props, body)
                        continue

                    if inflight.done.is_set():
                        self._finish_inflight(inflight)
                        inflight = None
                        continue

                    self._maybe_send_progress(inflight)
                    self._connection.process_data_events(time_limit=0.2)
            except KeyboardInterrupt:
                logger.info("TanukiWorker stopping (KeyboardInterrupt)")
                break
            except pika.exceptions.AMQPError:
                logger.exception("AMQP error; reconnecting soon")
                time.sleep(1.0)
            finally:
                self.close()

    def start(self, *, configure_logging: bool = True, log_level: int | str = "INFO") -> None:
        self.run(configure_logging=configure_logging, log_level=log_level)


def _coerce_log_level(level: int | str) -> int:
    if isinstance(level, int):
        return level
    name = level.strip().upper()
    mapping = logging.getLevelNamesMapping()
    if name not in mapping:
        raise ValueError(f"unknown log level: {level!r}")
    return int(mapping[name])


def _ensure_default_logging(level: int | str) -> None:
    target_level = _coerce_log_level(level)
    logging.getLogger("tanu").setLevel(target_level)

    root_logger = logging.getLogger()
    if root_logger.handlers:
        return
    logging.basicConfig(
        level=target_level,
        format="%(asctime)s %(levelname)s %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # Silence noisy pika internals by default (opt-in via user logging config).
    logging.getLogger("pika").setLevel(logging.WARNING)
    logging.getLogger("pika.adapters.blocking_connection").setLevel(logging.WARNING)
    logging.getLogger("pika.adapters.utils.connection_workflow").setLevel(logging.CRITICAL)
    logging.getLogger("pika.adapters.utils.io_services_utils").setLevel(logging.CRITICAL)



from __future__ import annotations

import base64
import dataclasses
import datetime as dt
import gzip
import importlib
import io
import json
import os
import pickle
import tempfile
from pathlib import Path
from typing import Any

from ..exceptions import TanukiParcelError
from ..parcel import Parcel, get_default_s3_store

_TYPE_KEY = "__tanu_type__"

_AUTO_PARCEL_THRESHOLD = int(os.getenv("TANU_AUTO_PARCEL_THRESHOLD", "10000"))
_NUMPY: Any | None = None
_PANDAS: Any | None = None


def _maybe_numpy() -> Any | None:
    global _NUMPY
    if _NUMPY is False:  # type: ignore[comparison-overlap]
        return None
    if _NUMPY is None:
        try:
            _NUMPY = importlib.import_module("numpy")
        except Exception:
            _NUMPY = False  # type: ignore[assignment]
            return None
    return _NUMPY


def _maybe_pandas() -> Any | None:
    global _PANDAS
    if _PANDAS is False:  # type: ignore[comparison-overlap]
        return None
    if _PANDAS is None:
        try:
            _PANDAS = importlib.import_module("pandas")
        except Exception:
            _PANDAS = False  # type: ignore[assignment]
            return None
    return _PANDAS


def _default(obj: Any) -> Any:
    if isinstance(obj, Parcel):
        meta = obj._ensure_remote()
        return {_TYPE_KEY: "parcel", "kind": "file", **meta}

    np = _maybe_numpy()
    if np is not None:
        if isinstance(obj, np.ndarray):
            if int(obj.size) > _AUTO_PARCEL_THRESHOLD:
                tmp_path: Path | None = None
                try:
                    with tempfile.NamedTemporaryFile(suffix=".npy", delete=False) as f:
                        tmp_path = Path(f.name)
                    np.save(tmp_path, obj, allow_pickle=False)
                    meta = get_default_s3_store().put_file(
                        tmp_path,
                        filename="ndarray.npy",
                        content_type="application/x-npy",
                    )
                finally:
                    if tmp_path is not None:
                        try:
                            tmp_path.unlink()
                        except Exception:
                            pass
                return {_TYPE_KEY: "parcel", "kind": "numpy.ndarray", "format": "npy", **meta}

            buf = io.BytesIO()
            np.save(buf, obj, allow_pickle=False)
            return {_TYPE_KEY: "ndarray", "format": "npy", "npy": buf.getvalue()}

        if isinstance(obj, np.generic):
            return obj.item()

    pd = _maybe_pandas()
    if pd is not None:
        if isinstance(obj, pd.DataFrame):
            if int(obj.size) > _AUTO_PARCEL_THRESHOLD:
                tmp_path = None
                try:
                    with tempfile.NamedTemporaryFile(suffix=".pkl.gz", delete=False) as f:
                        tmp_path = Path(f.name)
                    obj.to_pickle(tmp_path, compression="gzip")  # type: ignore[call-arg]
                    meta = get_default_s3_store().put_file(
                        tmp_path,
                        filename="dataframe.pkl.gz",
                        content_type="application/x-python-serialize",
                    )
                finally:
                    if tmp_path is not None:
                        try:
                            tmp_path.unlink()
                        except Exception:
                            pass
                return {_TYPE_KEY: "parcel", "kind": "pandas.DataFrame", "format": "pickle+gzip", **meta}

            raw = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
            return {_TYPE_KEY: "dataframe", "format": "pickle+gzip", "pickle_gzip": gzip.compress(raw)}

    if dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj)
    if isinstance(obj, dt.datetime):
        return {_TYPE_KEY: "datetime", "value": obj.isoformat()}
    if isinstance(obj, dt.date):
        return {_TYPE_KEY: "date", "value": obj.isoformat()}
    if isinstance(obj, Path):
        return {_TYPE_KEY: "path", "value": str(obj)}
    if isinstance(obj, bytes):
        return {_TYPE_KEY: "bytes", "base64": base64.b64encode(obj).decode("ascii")}
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


def _object_hook(obj: dict[str, Any]) -> Any:
    t = obj.get(_TYPE_KEY)
    if not t:
        return obj
    if t == "datetime":
        return dt.datetime.fromisoformat(str(obj["value"]))
    if t == "date":
        return dt.date.fromisoformat(str(obj["value"]))
    if t == "path":
        return Path(str(obj["value"]))
    if t == "bytes":
        return base64.b64decode(str(obj["base64"]))
    if t == "parcel":
        kind = str(obj.get("kind") or "file")
        bucket = str(obj.get("bucket") or "")
        key = str(obj.get("key") or "")
        uri = obj.get("uri")
        filename = obj.get("filename")
        content_type = obj.get("content_type")
        size = obj.get("size")
        sha256 = obj.get("sha256")
        fmt = str(obj.get("format") or "")

        if not bucket or not key:
            raise TanukiParcelError("invalid parcel reference (missing bucket/key)")

        if kind == "file":
            return Parcel(
                None,
                uri=str(uri) if uri is not None else None,
                bucket=bucket,
                key=key,
                filename=str(filename) if filename is not None else None,
                content_type=str(content_type) if content_type is not None else None,
                size=int(size) if size is not None else None,
                sha256=str(sha256) if sha256 is not None else None,
            )

        np = _maybe_numpy()
        pd = _maybe_pandas()
        store = get_default_s3_store()

        if kind == "numpy.ndarray":
            if np is None:
                raise TanukiParcelError("numpy is required to decode numpy.ndarray Parcel")
            suffix = ".npy" if not fmt or fmt == "npy" else ""
            tmp_path: Path | None = None
            try:
                with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as f:
                    tmp_path = Path(f.name)
                store.download_file(bucket=bucket, key=key, dst=tmp_path)
                return np.load(tmp_path, allow_pickle=False)
            finally:
                if tmp_path is not None:
                    try:
                        tmp_path.unlink()
                    except Exception:
                        pass

        if kind == "pandas.DataFrame":
            if pd is None:
                raise TanukiParcelError("pandas is required to decode pandas.DataFrame Parcel")
            tmp_path = None
            try:
                with tempfile.NamedTemporaryFile(suffix=".pkl.gz", delete=False) as f:
                    tmp_path = Path(f.name)
                store.download_file(bucket=bucket, key=key, dst=tmp_path)
                return pd.read_pickle(tmp_path, compression="gzip")  # type: ignore[call-arg]
            finally:
                if tmp_path is not None:
                    try:
                        tmp_path.unlink()
                    except Exception:
                        pass

        raise TanukiParcelError(f"unknown parcel kind: {kind}")

    if t == "ndarray":
        np = _maybe_numpy()
        if np is None:
            raise TanukiParcelError("numpy is required to decode numpy.ndarray")
        data = obj.get("npy")
        if not isinstance(data, (bytes, bytearray)):
            raise TanukiParcelError("invalid ndarray payload")
        return np.load(io.BytesIO(bytes(data)), allow_pickle=False)

    if t == "dataframe":
        pd = _maybe_pandas()
        if pd is None:
            raise TanukiParcelError("pandas is required to decode pandas.DataFrame")
        data = obj.get("pickle_gzip")
        if not isinstance(data, (bytes, bytearray)):
            raise TanukiParcelError("invalid dataframe payload")
        return pickle.loads(gzip.decompress(bytes(data)))
    return obj


def encode_json(payload: Any) -> bytes:
    return json.dumps(payload, default=_default, separators=(",", ":"), ensure_ascii=False).encode("utf-8")


def decode_json(data: bytes) -> Any:
    return json.loads(data.decode("utf-8"), object_hook=_object_hook)
